/* display.c */

/* This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


/* INCLUDES **************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <arpa/inet.h>
#include <pthread.h>

#include "X_window.h"
#include "lut.h"
#include "serial.h"

#include "plugin.h"
#include "demos/demos.h"

/* DEFINES ***************************/

#define USE_X11 1
#define USE_SERIAL 1
#define ECHOMAX 255


/* Window size: */
#define MYX (GRIDX*40)
#define MYY (GRIDY*40)


/* GLOBALS ***************************/


/* The LUTMAX'th element is not actually used. */
#define LUTMAX 255
lut_t lut[LUTMAX+1];


/* The frame: */
pel_t frame[GRIDSIZE];

/* One intended for displaying.  This is RGB */
unsigned char dispframe[GRIDX*GRIDY*3];


/* Quit flag: */
unsigned char quit_flag=0;

useconds_t sleeptime,elapsed;

/* Plugin list (needs to be null-terminated): */
#define MAXPLUGS 9
plugin_data_t *cur_plug=0,*plugin[MAXPLUGS+1]={
	&mat_plugin,
	&fire_plugin,
	&blob_plugin,
	&flash_plugin,
	&text_plugin,
	&debug_plugin,
	&evenodd_plugin,
	&interference_plugin,
	0,
};
/* This is set to the number of plugins in the plugin list (which needs to
 * be null-terminated).  DO NOT CALL ANYTHING TO DO WITH PLUGINS BEFORE
 * SETTING THIS:
 */
unsigned plugin_size=0;


/* Allow more than one plugin to be running at a time.  Plugins are not
 * written so that more than one copy can be run at a time, but different
 * plugins may be run just fine.  Slots are numbered 1 through MAXSLOTS (to
 * match up with the keys you press to select).  slot tells you what slot
 * the keys affect, and slots says which slot the corresponding plugin is
 * in (0 for none).
 */
#define MAXSLOTS 9
unsigned slot=1,slots[MAXPLUGS]={0,0,0,0,0,0,0,0,0};


/* Need to store a color index separately from the plugins.
 * For the moment, use just one.
 */
unsigned plug_col=0;

/* This keeps our audio buffer full while allowing us to preserve timing */
FILE *audio_in;
extern FILE *poopen();


/* CODE ******************************/


/* Gets audio and keeps it current in audio_buff */
void *audiothread(void *ptr){
  int i;
	/* popen creates a pipe so we can read the output
		 of the program we are invoking */
  if (!(audio_in = popen("pacat -r --format=u8 -d alsa_output.pci_8086_27d8_sound_card_0_alsa_playback_0.monitor", "r"))) {
  //if (!(audio_in = popen("arecord --format=U8 -c 2 -r 44100", "r"))) {
    exit(1);
  }

  /* read and fill the audio buffer */
  while (fgets(audio_buff, sizeof(audio_buff), audio_in) != NULL ) { }
  pclose(audio_in);
}

/* Returns the plugin index of the plugin in slot s (returns plugin_size
 * for no result).
 */
unsigned getplug(unsigned s){
	unsigned a;
	if(!s || s>MAXSLOTS)return plugin_size;
	for(a=0;a<plugin_size && slots[a]!=s;++a);return a;
}


/* Selects slot s.  This DOES NOT set the slot variable, but it does set
 * cur_plug to whatever plugin occupies this slot (null if none).
 */
void select_slot(unsigned s){
	unsigned a;
	if(s>MAXSLOTS || !s){cur_plug=0;return;}
	a=getplug(s);
	if(a<plugin_size && slots[a]==s)cur_plug=plugin[a];else cur_plug=0;
}


/* The following routines set stuff for whatever cur_plug points at, if
 * anything.
 */
void setcol(unsigned col){
	col%=LUTMAX;if(cur_plug)if(cur_plug->data){
		cur_plug->data->color0[0]=lut[col].a;
		cur_plug->data->color0[1]=lut[col].r;
		cur_plug->data->color0[2]=lut[col].g;
		cur_plug->data->color0[3]=lut[col].b;
	}
}
void setdir(unsigned dir){
	if(cur_plug)if(cur_plug->data)cur_plug->data->dir=dir;
}
void setdens(unsigned dens){
	if(cur_plug)if(cur_plug->data)cur_plug->data->density=dens;
}
void setspeed(unsigned speed){
	if(cur_plug)if(cur_plug->data)cur_plug->data->speed=speed;
}
void setfft_iris_ceil(double fft_iris_ceil){
	if(cur_plug)if(cur_plug->data)cur_plug->data->fft_iris_ceil=fft_iris_ceil;
}

/* This sets the current slot, if valid, to the given plugin (by index to
 * plugin list).  If plug is >= plugin_size, the slot is turned
 * off.  This sets cur_plug.
 */
void setplug(unsigned plug){
	unsigned a;
	if(!slot || slot>MAXSLOTS)return;
	/* Find which plugins, if any, occupy this slot, and disable them. */
	for(a=0;a<plugin_size;++a)if(slots[a]==slot)slots[a]=0;
	/* Now, if plug is valid, put it in this slot. */
	if(plug<plugin_size)slots[plug]=slot;
}


/* These are the same as above, but print out stuff. */
void setfft_iris_ceil2(double fft_iris_ceil){
	unsigned p;
	p=getplug(slot);if(p>=plugin_size){
		printf("Slot %u, no plugin: FFT iris ceiling not set.\n",slot);
		return;
	}
	/* Depend on shortcutting: */
	if(!cur_plug || !cur_plug->data)
		printf("Slot %u, plugin %u: no data, so FFT iris ceiling not set.\n",slot,p+1);
	setfft_iris_ceil(fft_iris_ceil);
	printf("Slot %u, plugin %u: FFT iris ceiling set to %f.\n",slot,p+1,fft_iris_ceil);

}
void setcol2(unsigned col){
	unsigned p;
	p=getplug(slot);if(p>=plugin_size){
		//printf("Slot %u, no plugin: color not set.\n",slot);
		return;
	}
	/* Depend on shortcutting: */
	if(!cur_plug || !cur_plug->data){
		//printf("Slot %u, plugin %u: no data, so color not set.\n",slot,p+1);
	}
	setcol(col);
	//printf("Slot %u, plugin %u: color set to %u.\n",slot,p+1,col);
}
void setdir2(unsigned dir){
	unsigned p;
	p=getplug(slot);if(p>=plugin_size){
		printf("Slot %u, no plugin: direction not set.\n",slot);
		return;
	}
	/* Depend on shortcutting: */
	if(!cur_plug || !cur_plug->data)
		printf("Slot %u, plugin %u: no data, so direction not set.\n",slot,p+1);
	setdir(dir);
	printf("Slot %u, plugin %u: direction set to %u.\n",slot,p+1,dir);
}
void setdens2(unsigned dens){
	unsigned p;
	p=getplug(slot);if(p>=plugin_size){
		printf("Slot %u, no plugin: density not set.\n",slot);
		return;
	}
	/* Depend on shortcutting: */
	if(!cur_plug || !cur_plug->data)
		printf("Slot %u, plugin %u: no data, so density not set.\n",slot,p+1);
	setdens(dens);
	printf("Slot %u, plugin %u: density set to %08X.\n",slot,p+1,dens);
}
void setspeed2(unsigned speed){
	unsigned p;
	p=getplug(slot);if(p>=plugin_size){
		printf("Slot %u, no plugin: speed not set.\n",slot);
		return;
	}
	/* Depend on shortcutting: */
	if(!cur_plug || !cur_plug->data)
		printf("Slot %u, plugin %u: no data, so speed not set.\n",slot,p+1);
	setspeed(speed);
	printf("Slot %u, plugin %u: speed set to %08X.\n",slot,p+1,speed);
}


/* Set up the colors the way I want: */
void initLUT(){
	lut_iter=LUTMAX;
#warning DEBUG: first channel was all 0
	lut_set(lut,  0,0x00,0x00,0xFF,0x00);lut_set(lut,LUTMAX,0x00,0x00,0xFF,0x00);
	lut_set(lut, 32,0xFF,0xFF,0xFF,0xFF);
	lut_set(lut, 64,0x00,0x00,0x00,0xDD);
	lut_set(lut, 96,0xF0,0x60,0x00,0x60);
	lut_set(lut,128,0x00,0xFF,0x00,0xFF);
	lut_set(lut,160,0x00,0xFF,0x00,0x00);
	lut_set(lut,192,0x00,0xFF,0x70,0x00);
	lut_set(lut,224,0x00,0xFF,0xFF,0x00);
	lut_fill(lut);
}


/* Handle keystrokes: */
void readkey(Window win,unsigned code,const XEvent *event,const X_window_t *Xwin){
	unsigned a;
	double b;
	if(event->type!=KeyPress)return;
	/* 'Q' quits: */
	if(code=='q' && event->xkey.state&(ControlMask|ShiftMask)){
		quit_flag=1;return;
	}
	/* raise or lower the FFT iris (low-pass filtering) */
	if(code=='o'){
		if(cur_plug && cur_plug->data)b=cur_plug->data->fft_iris_ceil;else b=0.0;
		if(b>0.0)b-=0.1;else b=0.0;
		setfft_iris_ceil2(b);
	}
	if(code=='p'){
		if(cur_plug && cur_plug->data)b=cur_plug->data->fft_iris_ceil;else b=0.0;
		if(b<1000.0)b+=0.1;else b=1000.0;
		setfft_iris_ceil2(b);
	}
	if(code=='i'){
		if(sleeptime<1000000)sleeptime+=100;
		printf("%d\n",sleeptime);
	}
	if(code=='u'){
		if(sleeptime>100)sleeptime-=100;
		printf("%d\n",sleeptime);
	}

	/* Keys on the left side of the keyboard set a direction: */
	if(code=='w')setdir2(0);
	if(code=='e')setdir2(1);
	if(code=='d')setdir2(2);
	if(code=='c')setdir2(3);
	if(code=='x')setdir2(4);
	if(code=='z')setdir2(5);
	if(code=='a')setdir2(6);
	if(code=='q')setdir2(7);
	/* 'j' and 'k' lower and raise the density of the current plugin. */
	if(code=='k'){
		if(cur_plug && cur_plug->data)a=cur_plug->data->density;else a=0;
		if(a<0xEEEEEEEE)a+=0x11111111;else a=0xFFFFFFFF;
		setdens2(a);
	}
	if(code=='j'){
		if(cur_plug && cur_plug->data)a=cur_plug->data->density;else a=0;
		if(a<0x11111111)a=0;else a-=0x11111111;
		setdens2(a);
	}
	/* Brackets increase and decrease the speed. */
	if(code==']'){
		if(cur_plug && cur_plug->data)a=cur_plug->data->speed;else a=0;
		if(a<0xEEEEEEEE)a+=0x11111111;else a=0xFFFFFFFF;
		setspeed2(a);
	}
	if(code=='['){
		if(cur_plug && cur_plug->data)a=cur_plug->data->speed;else a=0;
		if(a<0x11111111)a=0;else a-=0x11111111;
		setspeed2(a);
	}
	/* 'h' and 'l' cycle the color. */
	if(code=='h'){
		if(plug_col>0)--plug_col;else plug_col=LUTMAX-1;setcol2(plug_col);
	}
	if(code=='l'){
		if(plug_col<LUTMAX)++plug_col;else plug_col=0;setcol2(plug_col);
	}
	/* '0'-'9' change the plugin.  Hold down shift to instead select a slot. */
	if(code>'0' && code<='9'){
		if(event->xkey.state&ShiftMask){
			/* Shift means set slot: */
			slot=code-'0';select_slot(slot);
			a=getplug(slot);if(a>=plugin_size){
				printf("Slot %u has been selected (no plugin).\n",slot);
			}else{
				printf("Slot %u, plugin %u has been selected.\n",slot,a+1);
			}
		}else{
			/* No shift means set plugin: */
			setplug(code-'1');a=getplug(slot);
			if(a<plugin_size)printf("Slot %u: set to plugin %u.\n",slot,a+1);
			else printf("Slot %u: no plugin.\n",slot);
		}
	}
	/* '0' clears the current slot. */
	if(code=='0'){
		/* This means clear the slot. */
		setplug(plugin_size);
		printf("Slot %u has been turned off.\n",slot);
	}
}

void *mainthread(void *ptr){
	char serialport[255];
	unsigned a,b;
	struct timeval tv1,tv2;
	Window win;
	sleeptime=1000;

	/* need to specify what port to write to 
	if(argc==1){
		printf("Setting default serial port to /dev/ttyUSB0\n");
		strcpy(serialport,"/dev/ttyUSB0");
	}
	else{
		printf("Serial port set to %s\n",argv[1]);
		strcpy(serialport,argv[1]);
	}
	*/
	strcpy(serialport,"/dev/ttyUSB0");


	/* stuff for the socket */
  int sock;                        /* Socket */
  struct sockaddr_in echoServAddr; /* Local address */
  struct sockaddr_in echoClntAddr; /* Client address */
  unsigned int cliAddrLen;         /* Length of incoming message */
  char echoBuffer[ECHOMAX];        /* Buffer for echo string */
  unsigned short echoServPort;     /* Server port */
  int recvMsgSize;                 /* Size of received message */


	echoServPort = 8080;
	if ((sock = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP)) < 0){
		printf("socket() failed");
		return 1;
	}

  /* Construct local address structure */
  memset(&echoServAddr, 0, sizeof(echoServAddr));   /* Zero out structure */
  echoServAddr.sin_family = AF_INET;                /* Internet address family*/
  echoServAddr.sin_addr.s_addr = htonl(INADDR_ANY); /* Any incoming interface  */
  echoServAddr.sin_port = htons(echoServPort);      /* Local port */

  /* Bind to the local address */
  if (bind(sock, (struct sockaddr *) &echoServAddr, sizeof(echoServAddr)) < 0)
      printf("bind() failed");

	/* make our socket receive non-blocking */
	fcntl(sock,F_SETFL,O_NONBLOCK);

#ifdef USE_SERIAL
	if(serial_open(serialport)){
		printf("Error opening serial port (%s).\n",strerror(errno));
		return 1;
	}
	printf("Opened serial port.\n");
#endif

#ifdef USE_X11
	if(X_init()<0){
		printf("X error.\n");return 1;
	}
	printf("Connected to X server.\n");

	if((win=X_make_win(0,0,MYX,MYY,"Dance Floor",readkey,0))==None){
		printf("Could not make a window.\n");X_quit();serial_close();return 1;
	}
#endif
	printf("Created Disco Floor window.\n");

	initLUT();printf("Initialized LUT.\n");

	/* Count plugins. */
	for(plugin_size=0;plugin_size<MAXPLUGS && plugin[plugin_size];++plugin_size);
	/* Clear all slots, and initialize all plugins. */
	for(a=0;a<plugin_size;++a){
		slots[a]=0;cur_plug=plugin[a];if(cur_plug){
			setdir(0);setspeed(0x88888888);setdens(0x88888888);setcol(plug_col);
			/* FIXME: Actually give a random seed? */
			if(cur_plug->init)if(cur_plug->init(0)<0){
				printf("Error initializing plugin %u.\n",a);
				/* Just for the fun of it, attempt to end stuff: */
				if(a>0)do{
					--a;cur_plug=plugin[a];if(cur_plug->end)cur_plug->end();
				}while(a>0);
				X_quit();serial_close();return 1;
			}
		}
	}
	/* Enable the first plugin, in the first slot. */
	slot=1;setplug(0);select_slot(slot);
	/* Clear the grid frame: */
	memset(frame,0,GRIDSIZE);
	printf("Initialized %u plugins.\n",plugin_size);
	char parameter[ECHOMAX];
	char value[ECHOMAX];

	while(!quit_flag){
		/* this makes the colors automatically cycle */
		if(plug_col>0)--plug_col;else plug_col=LUTMAX-1;setcol2(plug_col);

		/* socket stuff first */
    /* Set the size of the in-out parameter */
     cliAddrLen = sizeof(echoClntAddr);

     /* Block until receive message from a client */
     if ((recvMsgSize = recvfrom(sock, echoBuffer, ECHOMAX, 0,
        (struct sockaddr *) &echoClntAddr, &cliAddrLen)) > 0) {
			/* null terminate the string */
			echoBuffer[recvMsgSize] = '\0';
			/* parameterize our input */
			sscanf(echoBuffer,"%s %s",parameter,value);
			if(!strncmp(parameter,"color",ECHOMAX)){ setcol2(atoi(value)); }
			if(!strncmp(parameter,"plugin",ECHOMAX)){ setplug(atoi(value)); }
			if(!strncmp(parameter,"direction",ECHOMAX)){ setdir2(atoi(value)); }
			if(!strncmp(parameter,"density",ECHOMAX)){ setdens2(atoi(value)); }
			if(!strncmp(parameter,"speed",ECHOMAX)){ setspeed2(atoi(value)); }

		 }

		/* Get the time: */
		if(gettimeofday(&tv1,0)<0){
			printf("Error getting current time (%s).\n",strerror(errno));
			for(a=0;a<plugin_size;++a){
				cur_plug=plugin[a];if(cur_plug->end)cur_plug->end();
				X_quit();serial_close();return 1;
			}
		}

		/* Run through all the slots, in order, and generate. */
		for(a=1;a<=MAXSLOTS;++a){
			select_slot(a);if(!cur_plug || !cur_plug->generate)continue;
			if(cur_plug->generate(frame)<0){
				printf("Error generating frame.\n");
				for(a=0;a<plugin_size;++a){
					cur_plug=plugin[a];if(cur_plug->end)cur_plug->end();
					X_quit();serial_close();return 1;
				}
			}
		}
		/* Return to the original slot. */
		select_slot(slot);

		/* Our pixel has 4 elements (RGBA), but the computer screen only has 3.
		 * Copy just the first 3 elements to get a rough estimate of what to
		 * display */
		for(a=0,b=0;a<GRIDSIZE;a+=PEL_NUMEL,b+=3){
			memcpy(dispframe+b,frame+a+1,3);
		}
#ifdef USE_X11
		if(X_image(win,(char *)dispframe,GRIDX,GRIDY,MYX,MYY)<0){
			printf("Error writing to window (window closed?).\n");
			for(a=0;a<plugin_size;++a){
				cur_plug=plugin[a];if(cur_plug->end)cur_plug->end();
			}
			X_quit();serial_close();return 0;
		}
#endif

#ifdef USE_SERIAL
		if(serial_frame(frame)<0){
			printf("Error writing to serial port (%s).\n",strerror(errno));
			for(a=0;a<plugin_size;++a){
				cur_plug=plugin[a];if(cur_plug->end)cur_plug->end();
			}
			X_quit();serial_close();return 1;
		}
#endif

#ifdef USE_X11
		while(X_handle(1,256));
#endif
		/* Get the time: */
		if(gettimeofday(&tv2,0)<0){
			printf("Error getting current time (%s).\n",strerror(errno));
			for(a=0;a<plugin_size;++a){
				cur_plug=plugin[a];if(cur_plug->end)cur_plug->end();
				X_quit();serial_close();return 1;
			}
		}
		/* How long I want between frames (us): */
		//sleeptime=100000;
		/* Determine how long to sleep: */
		elapsed=(tv2.tv_sec-tv1.tv_sec)*1000000+tv2.tv_usec-tv1.tv_usec;
		if(elapsed<sleeptime)usleep(sleeptime-elapsed);
	}
	printf("Time to quit.  Bye.\n");
	for(a=0;a<plugin_size;++a){
		cur_plug=plugin[a];if(cur_plug->end)cur_plug->end();
	}
	X_quit();serial_close();return 0;

}

int main(int argc,char **argv) {
	pthread_t audio_th, main_th;
	char *audio_th_msg = "Audio Thread";
	char *main_th_msg = "Main Thread";
	int iret1,iret2;

	iret1=pthread_create(&audio_th,NULL,audiothread,(void*)audio_th_msg);
	iret2=pthread_create(&main_th,NULL,mainthread,(void*)main_th_msg);
	pthread_join(audio_th,NULL);
	pthread_join(main_th,NULL);
}
