/* lut.c */

/* Some subroutines for color LUT manipulation.
 */

/* This defines an LUT structure. Essentially, all this provides is a
 * subroutine (setLUT(...)) that makes for a clean way to initialize
 * certain points in an LUT, and then a subroutine (makeLUT(...)) that will
 * linearly interpolate between the points you provide to make a nice,
 * smooth color change.
 */

/* You need to set LUT_ITER to the number of iterations with which you are
 * dealing. That way, the LUT routines will know not to overflow.
 * LUT's should have LUT_ITER+1 elements.
 */


#ifndef LUT_H
#define LUT_H


/* TYPES *****************************/


/* The basic LUT structure: */
typedef struct {
	/* Amber/alpha, Red, Green, Blue, Used */
  unsigned char a,r,g,b,u;
} lut_t;


/* EXTERNALS *************************/


/* The maximum size of all LUT's: */
extern unsigned lut_iter;


/* PROTOTYPES ************************/


/* Just to make the LUT initializer cleaner.
 * Checks for overflow -- if index>lut_iter, it sets lut_iter instead.
 */
void lut_set(lut_t *lut,unsigned index,unsigned a,unsigned r,unsigned g,unsigned b);


/* Fills in the LUT.
 * The first and last elements should be used before this is called,
 * or their values at the end will be unpredictable.
 */
void lut_fill(lut_t *lut);


#endif
