/* serial.c */

/* Three routines.
 * Call serial_open() to open a connection and set it up.
 * Call serial_frame() to write a frame.
 * Call serial_close() to close and release the serial port.
 */

/* Protocol is pretty simple.  Write a 0 byte to signal the first byte of
 * the frame.  Every thing else is two bytes per pixel, with 4 bits per
 * color channel (four channels).  Any 0 byte is incremented to a 1.
 */


#ifndef SERIAL_H
#define SERIAL_H


/* INCLUDES **************************/


#include "plugin.h"


/* PROTOTYPES ************************/


/* Opens the serial port in non-blocking mode, and sets it up
 * the way I want.  Give it the device name.
 * Returns 0 for success, -1 for error (and sets errno).
 */
int serial_open(const char *name);


/* Closes the serial output. */
void serial_close();


/* Writes the given frame (4 bytes per pixel) to the output.
 * First, waits for the last write to finish.
 * Returns 0 for success, -1 for error (sets errno).
 */
int serial_frame(pel_t *frame);


#endif
